using System.Collections;
using System.Collections.Generic;
using Unity.Collections;
using UnityEngine;
using Unity.Mathematics;
using Unity.Jobs;
using UnityEditor;
using UnityEngine.Profiling;

#region DOD
public enum DODWalkable : byte
{
    obstacle = 0,
    walkable = 1,
}

public struct Fg { public float f, g; }

[Unity.Burst.BurstCompile]
public struct GetDODPathJob : IJob
{
    [ReadOnly] public NativeArray<DODWalkable> allWalkable;
    [WriteOnly] public NativeList<int> pathIndexs;
    public float2 startPos, endPos, bottomLeftPos;
    public int mapMaxNodeLength;
    public float hMultiplier;
    public NativeArray<Fg> FG;
    public NativeArray<int2> nighborOffsetList;
    public NativeList<int> openList;
    public NativeHashSet<int> closeList;

    const float STRAIGHT_COST = 10;
    const float DIAGONAL_COST = 14; // STRAIGHT_COST * Math.Sqrt(2)

    public void Execute()
    {
        int startIndex  = GetDODIndex(startPos.x, startPos.y);
        int endIndex    = GetDODIndex(endPos.x, endPos.y);

        pathIndexs.Clear();

        int totalCount = this.allWalkable.Length;
        for (int i = 0; i < totalCount; i++)
            FG[i] = new Fg { g = float.MaxValue };

        NativeArray<int> previousIndex = new NativeArray<int>(totalCount, Allocator.Temp, NativeArrayOptions.UninitializedMemory);
        for (int i = 0; i < totalCount; i++)
            previousIndex[i] = -1;

        FG[startIndex] = new Fg { f = H(startPos, endPos), g = 0 };
        openList.Add(startIndex);

        float lengthInterval = mapMaxNodeLength - 1;
        float2 topRightPos =
            new float2(bottomLeftPos.x + lengthInterval,
                       bottomLeftPos.y + lengthInterval);

        int nightborOffsetListCount = nighborOffsetList.Length;
        while (
                openList.Length > 0
            &&  PopLowestFCostDODIndex(openList) is int currentIndex && currentIndex != endIndex
        )
        {
            closeList.Add(currentIndex);

            float currentG = FG[currentIndex].g;
            float2 currentPos = GetDODPosition(currentIndex);

            for (int i = 0; i < nightborOffsetListCount; i++)
            {
                float2 nighborPos = currentPos + nighborOffsetList[i];

                if (!InsideRange(nighborPos, topRightPos)) continue;

                int nightborIndex = GetDODIndex(nighborPos.x, nighborPos.y);

                if (closeList.Contains(nightborIndex)) continue;

                if (allWalkable[nightborIndex] == DODWalkable.obstacle) continue;

                float tempG = currentG + H(currentPos, nighborPos);
                if (tempG < FG[nightborIndex].g)
                {
                    previousIndex[nightborIndex] = currentIndex;
                    float tempH = H(nighborPos, endPos) * hMultiplier;
                    FG[nightborIndex] = new Fg
                    {
                        f = tempH + tempG,
                        g = tempG,
                    };

                   if (!openList.Contains(nightborIndex) &&
                       !closeList.Contains(nightborIndex))
                    {
                        openList.Add(nightborIndex);
                    }
                }
            }
        }

        if (previousIndex[endIndex] >= 0)
        {
            pathIndexs.Add(endIndex);
            int currentIndex = endIndex;
            while (     currentIndex != startIndex 
                    &&  previousIndex[currentIndex] >= 0)
            {
                currentIndex = previousIndex[currentIndex];
                pathIndexs.Add(currentIndex);
            }
        }
        else
        {
            pathIndexs.Clear();
        }
    }

    public int GetDODIndex(float xPos, float yPos, int mapMaxLength, Vector2 bottomLeft) =>
        (int)((xPos - bottomLeft.x) * mapMaxLength + (yPos - bottomLeft.y));

    public int GetDODIndex(float xPos, float yPos) =>
        (int)((xPos - bottomLeftPos.x) * mapMaxNodeLength + (yPos - bottomLeftPos.y));

    public float2 GetDODPosition(int index)
    {
        int row     = index % mapMaxNodeLength;
        int column  = index / mapMaxNodeLength;

        return new float2(bottomLeftPos.x + column,
                          bottomLeftPos.y + row);
    }

    private bool InsideRange(float2 position, float2 topRightPos)
        => !(position.x < bottomLeftPos.x   ||
             position.y < bottomLeftPos.y   ||
             position.x > topRightPos.x     ||
             position.y > topRightPos.y);

    private float H(float2 beginPos, float2 endPos)
    {
        float xDistance = math.abs(beginPos.x - endPos.x);
        float yDistance = math.abs(beginPos.y - endPos.y);
        float remaining = math.abs(xDistance - yDistance);
        return DIAGONAL_COST * math.min(xDistance, yDistance) + remaining * STRAIGHT_COST;
    }

    private int PopLowestFCostDODIndex(NativeList<int> openList)
    {
        int lowestFCostNodeLocation = -1;
        int lowestFIndex = -1;
        float lowestFCost = float.MaxValue;
        for (int i = 0; i < openList.Length; i++)
        {
            int nextFIndex = openList[i];
            float nextF = FG[nextFIndex].f;
            if (nextF < lowestFCost)
            {
                lowestFIndex = nextFIndex;
                lowestFCost = nextF;
                lowestFCostNodeLocation = i;
            }
        }

        openList.RemoveAtSwapBack(lowestFCostNodeLocation);

        return lowestFIndex;
    }
}

#endregion


public class SZM_PathFindingManager : CustomSingleton<SZM_PathFindingManager>
{
    public const float STRAIGHT_COST = 10;
    public const float DIAGONAL_COST = 14;

    protected override void Initialize()
    {
        #region DOD
#if UNITY_EDITOR
        RegisterEditorEvent();
#endif
        #endregion
    }

    #region DOD


    public GetDODPathJob pathJob;
    private SZM_MapData currentMapData = null;
    private DODWalkable[] currentWalkableData = null;
    private float _hMultiplier = 1.0f;
    private int2[] neighborList = new int2[8] 
    {
        new int2(-1, -1),
        new int2(0, -1),
        new int2(1, -1),
        new int2(-1, 0),
        new int2(1, 0),
        new int2(-1, 1),
        new int2(0, 1),
        new int2(1, 1)
    };
    public void ConvertDataToDODNode(SZM_MapData mapData)
    {
        if (currentMapData == mapData)
        {
            pathJob.openList.Clear();
            pathJob.closeList.Clear();
        }
        else
        {
            Dispose();

            int squareLength = SZM_TerrainMaker.squareLength;
            int mapMaxLength = squareLength * mapData.GetMapMaxLength();
            int totalNodeCount = (int)math.pow(mapMaxLength, 2);

            float halfToNodeCenter = SZM_TerrainMaker.halfSquareLength - 0.5f;
            float minNodeCenterX = mapData.GetMinXCenter() - halfToNodeCenter;
            float minNodeCenterY = mapData.GetMinYCenter() - halfToNodeCenter;

            Vector2 bottomLeftPos = new Vector2(minNodeCenterX, minNodeCenterY);

            currentWalkableData = new DODWalkable[totalNodeCount];
            foreach (KeyValuePair<int, SZM_TerrainDataCell> data in mapData.AllTerrainData)
            {
                int2 terrainCenterInt = data.Key.GetTerrainCenter();

                for (int i = 0; i < data.Value.walkableData.Length; i++)
                {
                    int row     = i % squareLength;
                    int column  = i / squareLength;
                    Vector2 nodeCenterPosition = new Vector2(terrainCenterInt.x - (halfToNodeCenter - column),
                                                             terrainCenterInt.y - (halfToNodeCenter - row));
                    int index = pathJob.GetDODIndex(nodeCenterPosition.x, nodeCenterPosition.y, mapMaxLength, bottomLeftPos);
                    currentWalkableData[index] = (DODWalkable)data.Value.walkableData[i];
                }
            }

            pathJob = new GetDODPathJob()
            {
                mapMaxNodeLength = mapMaxLength,
                allWalkable = new NativeArray<DODWalkable>(currentWalkableData, Allocator.Persistent),
                nighborOffsetList = new NativeArray<int2>(neighborList, Allocator.Persistent),
                openList = new NativeList<int>(totalNodeCount, Allocator.Persistent),
                closeList = new NativeHashSet<int>(totalNodeCount, Allocator.Persistent),
                FG = new NativeArray<Fg>(totalNodeCount, Allocator.Persistent, NativeArrayOptions.UninitializedMemory),
                pathIndexs = new NativeList<int>(Allocator.Persistent),
                bottomLeftPos = bottomLeftPos,
                hMultiplier = _hMultiplier,
            };
        }
    }

    public void CalculateDODPath(float2 beginPos, float2 endPos)
    {
        pathJob.startPos    = beginPos;
        pathJob.endPos      = endPos;

        JobHandle pathJobHandle = pathJob.Schedule();
        pathJobHandle.Complete();
    }

    private void Dispose()
    {
        if (pathJob.allWalkable.IsCreated)          pathJob.allWalkable.Dispose();
        if (pathJob.pathIndexs.IsCreated)           pathJob.pathIndexs.Dispose();
        if (pathJob.FG.IsCreated)                   pathJob.FG.Dispose();
        if (pathJob.nighborOffsetList.IsCreated)    pathJob.nighborOffsetList.Dispose();
        if (pathJob.openList.IsCreated)             pathJob.openList.Dispose();
        if (pathJob.closeList.IsCreated)            pathJob.closeList.Dispose();
    }







#if UNITY_EDITOR
    private void RegisterEditorEvent()
    {
        EditorApplication.playModeStateChanged += ModeChanged;
    }

    private void ModeChanged(PlayModeStateChange playModeState)
    {
        if (playModeState == PlayModeStateChange.EnteredEditMode)
        {
            Dispose();
            EditorApplication.playModeStateChanged -= ModeChanged;
        }
    }
#endif

    #endregion



    #region OOP
    public Dictionary<Vector2, SZM_PathNode> AllNodes { get; private set; } = new Dictionary<Vector2, SZM_PathNode>();

    public void ConvertDataToNode(SZM_MapData mapData)
    {
        AllNodes.Clear();

        int squareLength = SZM_TerrainMaker.squareLength;

        foreach (KeyValuePair<int, SZM_TerrainDataCell> data in mapData.AllTerrainData)
        {
            int2 terrainCenterInt = data.Key.GetTerrainCenter();
            Vector2 terrainCenter = new Vector2(terrainCenterInt.x, terrainCenterInt.y);

            for (int i = 0; i < data.Value.walkableData.Length; i++)
            {
                int row = i % squareLength;
                int column = i / squareLength;

                Vector2 bottomLeftPosition =
                    terrainCenter - new Vector2(SZM_TerrainMaker.halfSquareLength, SZM_TerrainMaker.halfSquareLength);

                Vector2 nodeCenterPosition =
                    new Vector2(bottomLeftPosition.x + column + 0.5f,
                                bottomLeftPosition.y + row + 0.5f);

                SZM_PathNode tempNode = new SZM_PathNode();
                tempNode.canWalk = data.Value.walkableData[i] == 1;
                tempNode.gCost = float.MaxValue;
                tempNode.CalculateFCost();
                tempNode.position = nodeCenterPosition;

                AllNodes.Add(nodeCenterPosition, tempNode);
            }
        }
    }

    public SZM_PathNode GetNode(Vector2 postion)
    {
        AllNodes.TryGetValue(postion, out SZM_PathNode retNode);
        return retNode;
    }

    public Stack<SZM_PathNode> GetPath(SZM_PathNode beginNode, SZM_PathNode endNode)
    {
        if (!beginNode.canWalk)
        {
            Debug.LogError("Begin Node not walkable!");
            return null;
        }

        Dictionary<Vector2, SZM_PathNode> openDic = new Dictionary<Vector2, SZM_PathNode>();
        Dictionary<Vector2, SZM_PathNode> closeDic = new Dictionary<Vector2, SZM_PathNode>();
        openDic.Add(beginNode.position, beginNode);

        beginNode.gCost = 0;
        beginNode.hCost = CalculateDistanceCost(beginNode, endNode);
        beginNode.CalculateFCost();


        while (openDic.Count > 0)
        {
            SZM_PathNode currentNode = GetLowestFCostPosition(openDic);
            if (currentNode.position == endNode.position)
                return CalculatePath(endNode);

            closeDic.Add(currentNode.position, openDic[currentNode.position]);
            openDic.Remove(currentNode.position);

            neighborNodeList.Clear();
            // Left, right, down, up
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Left));
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Right));
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Up));
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Down));
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Right_Up));
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Right_Down));
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Left_Up));
            TryAddNodeToList(neighborNodeList, currentNode.position.GetDirection(Direction.Left_Dowm));

            for (int i = 0; i < neighborNodeList.Count; i++)
            {
                if (closeDic.ContainsKey(neighborNodeList[i].position)) continue;

                float tempGCost = currentNode.gCost + CalculateDistanceCost(currentNode, neighborNodeList[i]);

                if (tempGCost < neighborNodeList[i].gCost)
                {
                    neighborNodeList[i].previousNode = currentNode;
                    neighborNodeList[i].gCost = tempGCost;
                    neighborNodeList[i].hCost = CalculateDistanceCost(neighborNodeList[i], endNode);
                    neighborNodeList[i].CalculateFCost();

                    openDic.TryAdd(neighborNodeList[i].position, neighborNodeList[i]);
                }
            }
        }

        return null;
    }

    private List<SZM_PathNode> neighborNodeList = new List<SZM_PathNode>(8);
    public void TryAddNodeToList(List<SZM_PathNode> targetList, Vector2 keyPosition)
    {
        if (AllNodes.TryGetValue(keyPosition, out SZM_PathNode node) &&
           node.canWalk)
        {
            targetList.Add(node);
        }
    }

    private Stack<SZM_PathNode> CalculatePath(SZM_PathNode endNode)
    {
        Stack<SZM_PathNode> path = new Stack<SZM_PathNode>();
        path.Push(endNode);
        SZM_PathNode currentNode = endNode;
        while (currentNode.previousNode != null)
        {
            path.Push(currentNode.previousNode);
            currentNode = currentNode.previousNode;
        }
        return path;
    }

    private SZM_PathNode GetLowestFCostPosition(Dictionary<Vector2, SZM_PathNode> dic)
    {
        SZM_PathNode retNode = null;
        float lowestFCost = float.MaxValue;
        foreach (KeyValuePair<Vector2, SZM_PathNode> data in dic)
        {
            if (data.Value.fCost < lowestFCost)
            {
                lowestFCost = data.Value.fCost;
                retNode = data.Value;
            }
        }
        return retNode;
    }

    private float CalculateDistanceCost(SZM_PathNode beginNode, SZM_PathNode endNode)
    {
        float xDistance = Mathf.Abs(beginNode.position.x - endNode.position.x);
        float yDistance = Mathf.Abs(beginNode.position.y - endNode.position.y);
        float remaining = Mathf.Abs(xDistance - yDistance);
        return DIAGONAL_COST * Mathf.Min(xDistance, yDistance) + remaining * STRAIGHT_COST;
    }

    #endregion
}

public static class SZM_PathFIndingExtension
{
    public static Vector2 GetDirection(this Vector2 centerPositin, Direction direction)
    {
        return new Vector2(centerPositin.x + ((direction & Direction.Right) > 0 ? 1 : 0) - ((direction & Direction.Left) > 0 ? 1 : 0),
                           centerPositin.y + ((direction & Direction.Up) > 0 ? 1 : 0) - ((direction & Direction.Down) > 0 ? 1 : 0));
    }

    public static SZM_PathNode GetNode(this Vector2 postion)
    {
        return SZM_PathFindingManager.Instance.GetNode(postion);
    }
}


public enum Direction
{
    Up = 0x00000001,
    Right = 0x00000002,
    Down = 0x00000004,
    Left = 0x00000008,
    Right_Up = 0x00000003,
    Right_Down = 0x00000006,
    Left_Dowm = 0x0000000c,
    Left_Up = 0x00000009,
}